/**
 * Mitsubishi Projektor-Steuerung
 * 
 * Verbindet sich via Ethernet zu einem Mitsubishi-Projektor im lokalen
 * Netzwerk und stellt Steuerfunktionen für diesen bereit.
 * 
 * Getestete Projektoren:
 * - Mitsubishi HC7800D
 * 
 * @author Bert Kößler <bert@heimkino-praxis.com>
 * @copyright © 2015 Bert Kößler
 */

var _mitsubishi = exports;

exports.put = function (command, settings, callback)
{
	this.send(command, settings, callback);
};

exports.get = function (command, settings, callback)
{
	this.send(command, settings, callback);
};

exports.send = function (command, settings, callback)
{
	var data = {
		ip: settings.ip,
		port: settings.port,
		password: settings.password,
		command: command
	};
	
	this.executeCommand(data, callback);
};

exports.executeCommand = function (data, callback)
{
	var net = require("net");
	var crypto = require("crypto");
	
	var host = data.ip || "192.168.0.1";
	var port = data.port || 63007;
	
	if (!Array.isArray(data.command)) {
		data.command = [ data.command ];
	}

	var socket = new net.Socket();
	
	socket.connect(port, host, function() {
		socket.write("$AK");
		socket.write("\r");
	});
	
	socket.setTimeout(2000, function () {
		socket.end();
		socket.destroy();
		
		if (callback)
		{
			callback({
				error: "Device not responding / Command currently unavailable"
			});
		}
	});
	
	socket.on("error", function (error)
	{
		socket.end();
		socket.destroy();
		
		if (callback)
		{
			callback({
				error: error
			});
		}
	});
	
	var digest = "";
	var messages = new Array();
	
	socket.on("data", function(message)
	{
		message = message.toString("utf8").trim();
		
		if (message.substring(0, 3) == "$AK")
		{
			var pwtoken = message.substring(3, 11);
			
			digest = crypto.createHash("md5")
				.update(pwtoken + data.password)
				.digest("hex");
		}
		else {
			messages.push(message);
		}
		
		if (data.command.length > 0)
		{
			socket.write(digest);
			socket.write(data.command.shift());
			socket.write("\r");
			
			return;
		}
		
		var resultCode = message.length > 0 ? 0 : 1;
		
		if (callback)
		{
			var responseObject = {
				result: messages.length == 1 ? messages[0] : messages,
				code: resultCode,
				message: resultCode == 0 ? "OK" : "Unknown result"
			};
			
			if (resultCode > 0)
			{
				console.log(responseObject.result);
				responseObject.error = "Unknown result";
			}
			
			callback(responseObject);
		}
		
		socket.destroy();
	});

	socket.on("close", function() {
		//console.log("Mitsubishi: Connection closed");
	});
};

exports.parseResponse =
{
	asString: function (result, command)
	{
		if (!result) return undefined;
		return result.substring(result.indexOf(command) + command.length);
	},
	asInteger: function (result, command)
	{
		if (!result) return undefined;
		return parseInt(_mitsubishi.parseResponse.asString(result, command));
	},
	asBoolean: function (result, command)
	{
		if (!result) return undefined;
		return _mitsubishi.parseResponse.asString(result, command) == "1";
	}
};

exports.Projector = function (options)
{
	var _self = this;
	
	var settings = {
		id: "Mitsubishi.Projector",
		ip: "192.168.0.1",
		port: 63007,
		password: "admin"
	};
	
	this.setup = function (data)
	{
		for (var property in data) {
			settings[property] = data[property];
		}
	};
	
	if (options) {
		this.setup(options);
	}
	
	this.getId = function ()
	{
		return settings.id;
	};
	
	this.getIp = function ()
	{
		return settings.ip;
	};
	
	this.getPort = function ()
	{
		return settings.port;
	};
	
	this.getPower = function (callback)
	{
		this.getStatus(function (value) {
			if (value > 3) value = 2;
			callback(value);
		});
	};
	
	this.setPower = function (value, callback)
	{
		switch (value)
		{
			case 1:
			case "On":
			case true:
				_mitsubishi.put("00!", settings, callback);
				return;
				
			case 0:
			case "Off":
			case false:
				_mitsubishi.put('00"', settings, callback);
				return;
				
			default:
				_self.getPower(function (power) {
					if (power == 2) {
						_self.setPower(false, callback);
					} else if (power == 0) {
						_self.setPower(true, callback);
					}
				});
				return;
		}
	};
	
	this.getStatus = function (callback)
	{
		_mitsubishi.get("00vST", settings, function (response) {
			callback(_mitsubishi.parseResponse.asInteger(response.result, "00vST"));
		});
	};
	
	this.getSignalSupplied = function (callback)
	{
		_mitsubishi.get("00vSM", settings, function (response) {
			callback(_mitsubishi.parseResponse.asInteger(response.result, "00vSM"));
		});
	};
	
	this.getInputSources = function (callback)
	{
		_mitsubishi.get("00vSOUCE", settings, function (response) {
			var value = _mitsubishi.parseResponse.asString(response.result, "00vSOUCE");
			callback(value.trim().split(" "));
		});
	};
	
	this.getInput = function (callback)
	{
		_mitsubishi.get("00vI", settings, function (response) {
			callback(_mitsubishi.parseResponse.asString(response.result, "00vI"));
		});
	};
	
	this.setInput = function (value, callback)
	{
		_mitsubishi.put("_" + value, settings, callback);
	};
	
	// TODO: prüfen, ob dieser Befehl unterstützt wird
	this.toggleAudioVideoMute = function (callback)
	{
		_mitsubishi.put("00ra6", settings, callback);
	};
	
	this.toggleMenu = function (callback)
	{
		_mitsubishi.put("00r54", settings, callback);
	};
	
	this.controlPadUp = function (callback)
	{
		_mitsubishi.put("00r53", settings, callback);
	};
	
	this.controlPadDown = function (callback)
	{
		_mitsubishi.put("00r2b", settings, callback);
	};
	
	this.controlPadLeft = function (callback)
	{
		_mitsubishi.put("00r4f", settings, callback);
	};
	
	this.controlPadRight = function (callback)
	{
		_mitsubishi.put("00r59", settings, callback);
	};
	
	this.controlPadEnter = function (callback)
	{
		_mitsubishi.put("00r10", settings, callback);
	};
	
	// TODO: Auslesen von Einstellungen funktioniert noch nicht immer / nur bei manchen Einstellungen
	// TODO: setPictureMode scheint nicht zu funktionieren
	this.getPictureMode = function (callback)
	{
		_mitsubishi.get("00PM", settings, function (response) {
			callback(_mitsubishi.parseResponse.asInteger(response.result, "00PM"));
		});
	};
	
	this.setPictureMode = function (value, callback)
	{
		value = Math.min(8, Math.max(0, parseInt(value)));
		_mitsubishi.put("00PM" + value, settings, callback);
	};
	
	this.setAVMemory = function (value, callback)
	{
		value = Math.min(3, Math.max(1, parseInt(value)));
		_mitsubishi.put("00re" + (value + 3), settings, callback);
	};
	
	this.get3D = function (callback)
	{
		_mitsubishi.get("00TDE", settings, function (response) {
			callback(_mitsubishi.parseResponse.asInteger(response.result, "00TDE"));
		});
	};
	
	this.set3D = function (value, callback)
	{
		switch (value)
		{
			case 1:
			case "On":
			case true:
				_mitsubishi.put("00TDE1", settings, callback);
				return;
				
			case 0:
			case "Off":
			case false:
				_mitsubishi.put("00TDE0", settings, callback);
				return;
				
			default:
				_self.get3D(function (v) {
					_self.set3D(!v, callback);
				});
				return;
		}
	};
	
	this.get3DMode = function (callback)
	{
		_mitsubishi.get("00TDM", settings, function (response) {
			callback(_mitsubishi.parseResponse.asInteger(response.result, "00TDM"));
		});
	};
	
	this.set3DMode = function (value, callback)
	{
		if (typeof value == "string") {
			value = value.toLowerCase();
		}
		
		switch (value)
		{
			case 1:
			case "fp":
			case "frame packing":
				_mitsubishi.put("00TDM1", settings, callback);
				return;
				
			case 2:
			case "sbs":
			case "side by side":
				_mitsubishi.put("00TDM2", settings, callback);
				return;
				
			case 3:
			case "tb":
			case "top/bottom":
			case "top and bottom":
				_mitsubishi.put("00TDM3", settings, callback);
				return;
				
			case 0:
			case "auto":
			default:
				_mitsubishi.put("00TDM0", settings, callback);
				return;
		}
	};
	
	this.get3DSyncReverse = function (callback)
	{
		_mitsubishi.get("00TDS", settings, function (response) {
			callback(_mitsubishi.parseResponse.asInteger(response.result, "00TDS"));
		});
	};
	
	this.set3DSyncReverse = function (value, callback)
	{
		switch (value)
		{
			case 1:
			case "On":
			case true:
				_mitsubishi.put("00TDS1", settings, callback);
				return;
				
			case 0:
			case "Off":
			case false:
				_mitsubishi.put("00TDS0", settings, callback);
				return;
				
			default:
				_self.get3DSyncReverse(function (v) {
					_self.set3DSyncReverse(!v, callback);
				});
				return;
		}
	};
	
	this.getTestPattern = function (callback)
	{
		_mitsubishi.get("00TP", settings, function (response) {
			callback(_mitsubishi.parseResponse.asInteger(response.result, "00TP"));
		});
	};
	
	this.setTestPattern = function (value, callback)
	{
		if (typeof value == "string") {
			value = value.toLowerCase();
		}
		
		switch (value)
		{
			case 1:
			case "cross hatch":
				_mitsubishi.put("00TP1", settings, callback);
				return;
				
			case 2:
			case "white":
				_mitsubishi.put("00TP2", settings, callback);
				return;
				
			case 3:
			case "black":
				_mitsubishi.put("00TP3", settings, callback);
				return;
				
			case 0:
			case "off":
			default:
				_mitsubishi.put("00TP0", settings, callback);
				return;
		}
	};
	
	/**
	 * Zeigt einen Text auf dem Bildschirm an.
	 * 
	 * Der Text kann maximal 23 Zeichen lang sein. Gültige Zeichen sind Groß- und
	 * Kleinbuchstaben, Zahlen, Leerzeichen sowie !?,/&#@-+=()":;[]$<>.
	 * 
	 * @param String text Die Nachricht, die angezeigt werden soll.
	 * @param int duration Die Dauer in Millisekunden, für die der Text angezeigt werden soll, oder 0 für unbegrenzte Dauer.
	 * @param object userFlags Ein Objekt, das das Aussehen der Nachricht beeinflusst (vergleiche defaultFlags).
	 * @param function callback Eine Callback-Funktion, die aufgerufen wird, wenn die Nachricht angezeigt wurde.
	 * @param function clearCallback Eine Callback-Funktion, die aufgerufen wird, wenn die Nachricht wieder ausgeblendet wurde.
	 */
	this.showMessage = function (text, duration, userFlags, callback, clearCallback)
	{
		var duration = duration || 0;
		var userFlags = userFlags || {};
		
		var defaultFlags = {
			"italic": false,
			"underline": false,
			"textColor": "white",
			"borderColor": "white",
			"borderStyle": "solid",
			"blink": false,
			"align": "center",
			"valign": "center"
		};
		
		var flags = {
			"italic": userFlags.italic || defaultFlags.italic,
			"underline": userFlags.underline || defaultFlags.underline,
			"textColor": userFlags.textColor || defaultFlags.textColor,
			"borderColor": userFlags.borderColor || defaultFlags.borderColor,
			"borderStyle": userFlags.borderStyle || defaultFlags.borderStyle,
			"blink": userFlags.blink || defaultFlags.blink,
			"align": userFlags.align || defaultFlags.align,
			"valign": userFlags.valign || defaultFlags.valign
		};
		
		var flagValues = {
			// flag 1, char 1-2: font style
			"font": {
				"normal": "00",
				"italic": "01",
				"underline": "02",
				"italicUnderline": "03"
			},
			// flag 1, char 3: font color
			// flag 1, char 4: border color
			"color": {
				"blue": "1",
				"green": "2",
				"cyan": "3",
				"red": "4",
				"magenta": "5",
				"yellow": "6",
				"white": "7"
			},
			// flag 1, char 5: border style
			"border": {
				"none": "0",
				"solid": "1",
				"double": "2",
				"dotted": "3"
			},
			// flag 1, char 6: blink
			"blink": {
				"off": "0",
				"on": "1"
			},
			// flag 2, char 1-2: vertical position
			// flag 2, char 3-4: horizontal position
			"position": {
				"vertical": {
					"top": "11",
					"center": "15",
					"bottom": "19"
				},
				"horizontal": {
					"left": "01",
					"center": "07",
					"right": "19"
				}
			}
		};
		
		if (text.length > 23) {
			text = text.substring(0, 22);
		}
		
		var flag1 = "";
		
		if (flags.italic === true && flags.underline === true) {
			flag1 += flagValues.font.italicUnderline;
		} else if (flags.underline === true) {
			flag1 += flagValues.font.underline;
		} else if (flags.italic === true) {
			flag1 += flagValues.font.italic;
		} else {
			flag1 += flagValues.font.normal;
		}
		
		if (flags.textColor in flagValues.color) {
			flag1 += flagValues.color[flags.textColor];
		} else {
			flag1 += flagValues.color.white;
		}
		
		if (flags.borderColor in flagValues.color) {
			flag1 += flagValues.color[flags.borderColor];
		} else {
			flag1 += flagValues.color.white;
		}
		
		if (flags.borderStyle in flagValues.border) {
			flag1 += flagValues.border[flags.borderStyle];
		} else {
			flag1 += flagValues.border.solid;
		}
		
		if (flags.blink === true) {
			flag1 += flagValues.blink.on;
		} else {
			flag1 += flagValues.blink.off;
		}
		
		var flag2 = "";
		
		if (flags.valign in flagValues.position.vertical) {
			flag2 += flagValues.position.vertical[flags.valign];
		} else {
			flag2 += flagValues.position.vertical.center;
		}
		
		if (flags.align in flagValues.position.horizontal) {
			flag2 += flagValues.position.horizontal[flags.align];
		} else {
			flag2 += flagValues.position.horizontal.center;
		}
		
		_mitsubishi.put(["00&" + flag1, "00&" + flag2 + text, "00&21"], settings, function ()
		{
			if (callback) {
				callback();
			}
			
			if (duration <= 0) {
				return;
			}
			
			setTimeout(function () {
				_self.clearMessage(clearCallback);
			}, duration);
		});
	};
	
	/**
	 * Löscht eine angezeigte Nachricht vom Bildschirm.
	 * 
	 * @param function callback Eine Callback-Funktion, die aufgerufen wird, wenn die Nachricht gelöscht wurde.
	 */
	this.clearMessage = function (callback)
	{
		_mitsubishi.put("00&20", settings, callback);
	};
};